classdef HeaveModel < handle
    
    properties
        param;
        idx;
        
        % properties for the boundary conditions
        continuous_circulation = false;
        q_pump = @(t) 0;
        choke_opening = @(t) 1;  % \in [0,1], 0 = closed, 1 = open
        choke_mode = 'open';
        q_backpressure_pump = @(t) 0;
        v_rig = @(t) 0;
        qbit_des = @(t) 0;
        states_simulink = 0;
        heave_comp = @(t) 0;
        
        
        u_hl_end = 1;
        
        % HeaveLock
        useHeaveLock = false;
        useSimulink = false;
        positionHeaveLock = 1; 
        
        dim;
        
        % tracking variables
        track_pi_top = true;
        pi_top_tracked = [];
        t_tracked = [];
        q_bit_tracked = [];
        q_bha_tracked = [];
        q_HL_tracked = [];   
        q_des_tracked = [];
        u_tracked = [];
        
        debug = [];
        
        % coupling inner and annular pressure via pipe deformations
        coupling_pi_pa = true;
        
        % way to compute pressure drop in bit/ around bha
        pressuredropexceldata = true;
        
        % lumped pipe model (only one degree of freedom) or not
        lump_pipe = false;
        
        switch_off_i = false;
        
        switch_off_a = false;   
    end
    
    methods
        
        function this = HeaveModel(param)
            this.param = param;
            this.init_idx;
        end
        
        
        function [t,x] = simulate(this, t, x0)
            if this.useSimulink
                while strcmp('paused',get_param('getFeedforwardSignal','SimulationStatus')) % check if model is already compiled
                    getFeedforwardSignal([],[],[],'term');
                end
                [sys,~,~,~] = getFeedforwardSignal([],[],[],'compile');
                this.states_simulink = sys(1);
                set_param('getFeedforwardSignal', 'InitInArrayFormatMsg', 'None')
            end
            this.init_idx;
            
            if nargin < 3
                if false;this.continuous_circulation;
                    x0 = this.equilibrate;
                else
                    x0 = this.getX0;
                end
            end
            
            % reset tracked variables
            this.t_tracked = [];
            this.q_bit_tracked = [];
            this.q_bha_tracked = [];
            this.q_des_tracked = [];
            this.u_tracked = [];
            
            % simulate
            options = odeset('RelTol',1e-3);
            %[t, x] = ode23(@this.computeRHS, t ,x0);
            [t, x] = ode23t(@this.computeRHS, t ,x0,options);
            if this.useSimulink
                getFeedforwardSignal([],[],[],'term');
            end
            x = x';
            t = t';
            %[t, x] = ode113(@this.computeRHS, t ,x0);
            
            % track variables
            [t_tracked,indx,~] = unique(this.t_tracked);
            if this.track_pi_top
                x(this.idx.p_i(end),:) = interp1(t_tracked, this.pi_top_tracked(indx),t);
            end
            this.q_bit_tracked = interp1(t_tracked, this.q_bit_tracked(indx),t);
            this.q_bha_tracked = interp1(t_tracked, this.q_bha_tracked(indx),t);
            this.q_HL_tracked = interp1(t_tracked, this.q_HL_tracked(indx),t);
            this.q_des_tracked = interp1(t_tracked, this.q_des_tracked(indx),t);
            this.u_tracked = interp1(t_tracked, this.u_tracked(indx),t);
            
        end
        
        
        
        function dx = computeRHS(this, t,x)
            % t time, x state
            
            % parameters
            p = this.param;
            idx = this.idx;
            deltaX = p.L/p.N;
            g = 9.81;
            
            % select states
            p_i = x(idx.p_i);
            q_i = x(idx.q_i);
            p_p = x(idx.p_p);
            v_p = x(idx.v_p);
            p_a = x(idx.p_a);
            q_a = x(idx.q_a);
            v_BHA = x(idx.v_BHA);
            p_bh = x(idx.p_bh);
            x_bit = x(idx.x_bit);
            q_bit = x(idx.q_bit);
            q_BHA = x(idx.q_BHA);
            fluid_level_i = x(idx.fluid_level_i);  % fluid level in pipe
            q_des  = 0;
            
            %ToDo: keep or remove?
            x_p = x(idx.x_p);


            %% boundary/coupling conditions
            A_well = p.A_i + p.A_p + p.A_a;
            
            % pressure drop and flow around BHA (in annulus)
            delta_p = p_bh - p_a(1);   % positive pressure difference <=> flow upwards
            if ~this.pressuredropexceldata
                % from Imsland's report: delta_p = p.rho*q_bha^2/(2*C_d^2*A_gap^2);
                C_d = 0.95;
                A_gap = A_well - p.A_bha;
                hlp = delta_p*2*C_d^2*A_gap^2/p.rho;
                q_bha = sign(delta_p)*sqrt(abs(hlp));
            else
                % ToDo: remove, factor 100 in q wrong: alternative: delta_p = 0.03*q^1.25 (in bar, fitted to pressure loss data
                
                % alternative: delta_p = 169*q^1.25 (in bar, fitted to pressure loss data
                % in excel files X5 ARC PowerPulse
                hlp = delta_p/169/1e5;
                q_bha = sign(hlp)*(abs(hlp))^0.8;
            end
            
            
            if this.continuous_circulation
            % pressure drop and flow through bit/BHA (inside drill string)
            delta_p = p_i(1) - p_bh;
                if delta_p > 0
                    if ~this.pressuredropexceldata
                        % Imsland's report:
                        C_d = 0.95;
                        A_bit_nozzles = p.A_p;  % ToDo
                        q_bit = -sqrt(delta_p*2*C_d^2*A_bit_nozzles^2/p.rho);
                    else
                        % wrong:  alternative: delta_p = 0.09*q^2 (fitted to pressure loss data
                        % alternative: delta_p = 90000*q^2 (fitted to pressure loss data
                        % in excel files X5 ARC PowerPulse
                        q_bit = -sqrt(delta_p/90000/1e5);
                    end
                else % non return valve closed
                    q_bit = 0;
                end
            else
                q_bit = 0;
            end
            
%             v_BHA = this.v_rig(t);
            
            if this.useSimulink
                x_simulink = x(idx.simulink);
                dqbit = getFeedforwardSignal(t,x_simulink,v_BHA,'outputs');
                dx_simulink = getFeedforwardSignal(t,x_simulink,v_BHA,'derivs');
            end
            
            %%%%%%%%%%%%%
            % HeaveLock %
            %%%%%%%%%%%%%
            if this.useSimulink
                q_HeaveLock = this.q_pump(t)-dqbit;
            end
%               q_bit = this.q_pump(t)-qbit_des(t);
            u_hl = p.u_hl_init;
            if this.useHeaveLock  
                delta_p_hl = p_i(this.positionHeaveLock+1) - p_i(this.positionHeaveLock) + p.g*p.rho*(deltaX-1);
%                 q_i(this.positionHeaveLock) = this.q_pump(t)-v_BHA*(p.A_p+p.A_bit);%q_HeaveLock;
                if     t >= p.t_hl1 && t < p.t_hl1 + p.delta_t_hl1
                    % Ramp to initial position
                    u_hl = (p.u_hl0-p.u_hl_init)/p.delta_t_hl1*(t-p.t_hl1) + p.u_hl_init;
                    %u_hl = (p.u_hl0-1)/p.delta_t_hl1*(t-p.t_hl1) + 1;
                    q_des = 0;
%                     disp(1)
                elseif t >= p.t_hl1 + p.delta_t_hl1 && t < p.t_hl_start
                    % Wait for start "signal"
                    u_hl = p.u_hl0;
                    q_des = 0;
%                     disp(2)
                elseif t >= p.t_hl_start && t < p.t_hl_end
                    % Active control mode
                    q_des = this.q_pump(t)-v_BHA*1*(p.A_p+p.A_bit);
                    if delta_p_hl < 0
                        disp('negative dp_hl')
                        stop
                    end
                    u_hl = -q_des/(p.k_hl*sqrt(delta_p_hl));
                    if u_hl > 1
                        u_hl = 1;
                    elseif u_hl < p.u_min
                        u_hl = p.u_min;
                    end
%                     disp(3)
                elseif t >= p.t_hl_end && t < p.t_hl2
                    % End active control mode, hold last opening
                    u_hl = this.u_tracked(end);
                    this.u_hl_end = u_hl;
                    q_des = 0;
%                     disp(4)
                elseif t >= p.t_hl2 && t < p.t_hl2 + p.delta_t_hl2
                    % Ramp to fully open
                    u_hl = (p.u_hl_init-this.u_hl_end)/p.delta_t_hl2*(t-p.t_hl2) + this.u_hl_end;
                    q_des = 0;
%                     disp(5)
                else
                    % Inactive
                    u_hl = p.u_hl_init;
                    q_des = 0;
%                     disp(6)
                end
                q_i(this.positionHeaveLock) = -p.k_hl*u_hl*sqrt(delta_p_hl);                
            end
            
            % ToDo: close drillstring at bottom
            %q_bit = 0;
            
            
            % pressure drop and flow through choke - ToDo, make q_choke a state??
            if strcmp(this.choke_mode, 'open') 
                q_choke = q_a(end);
            elseif strcmp(this.choke_mode, 'closed') 
                q_choke = 0;
            elseif strcmp(this.choke_mode, 'impedance_matching') 
                p0 = 1e5;
                delta_p = p_a(end)-p0;
                q_choke = 1/sqrt(p.beta_a*p.rho)*p.A_a*delta_p;
            elseif strcmp(this.choke_mode, 'dynamic') 
                % ToDo: choke equation
                %error('choke equation for mode ''dynamic'' is not implemented yet')
                p0 = 1e5;
                delta_p = p_a(end)-p0; %-0*0.5*deltaX*p.g*p.rho;
                
%                 C_d = 0.95;
%                 A_choke = 5*p.A_p;  % ToDo: load from param or so
%                 A_choke = A_choke*this.choke_opening(t);
                
                if delta_p > 0
%                     q_choke = 1e-5*2*sqrt(delta_p);
                    % q_choke = 2*1e-5/2/sqrt(2.3e5)*p_a(end);
                    q_choke = p.A_choke*this.choke_opening(t)*p.C_d*sqrt(delta_p*2/p.rho);
                else
                    q_choke = 0;
                end 
            else
                error('unknown choke mode ''%s''', this.choke_mode)
            end
            
            q_choke = q_choke + this.q_backpressure_pump(t);   % ToDo
            
            % flow into drillstring topside
            if this.continuous_circulation
                if t > p.t_p1 && t < p.t_p1+p.delta_t_p1
                    q_pump = this.q_pump(t);
                    
                end
                q_pump = this.q_pump(t);
            else
                %if (p_i(end) >= 1e5) && (q_i(end) > 0)
                %if (p_i(end) >= 0.5*deltaX*g*p.rho)
                %if (p_i(end) >= 0.5*deltaX*g*p.rho) && (q_i(end) > 0)
                if fluid_level_i >= p.L
                    q_pump = q_i(end);
                else
                    q_pump = 0;
                end
                %q_pump = 0;       
            end
            
            
            q_a = [q_bha+v_BHA*(p.A_bha-p.A_i-p.A_p)*1+0*(p.A_a)*v_BHA; q_a; q_choke];
            v_p = [v_BHA; v_p; this.v_rig(t)];
            q_i = [q_bit+(p.A_i-p.A_bit)*v_BHA; q_i; q_pump];
            
            
            %% x-derivatives
            
            dqa = (q_a(2:end)-q_a(1:end-1))/deltaX;
            
            dpa = (p_a(2:end)-p_a(1:end-1))/deltaX;
            dvp = (v_p(2:end)-v_p(1:end-1))/deltaX;
            dpp = (p_p(2:end)-p_p(1:end-1))/deltaX;
            dqi = (q_i(2:end)-q_i(1:end-1))/deltaX;
            dpi = (p_i(2:end)-p_i(1:end-1))/deltaX;

            
            %% friction & gravity terms
            [Fi, Fa, Fp] = this.evaluateFriction(q_i(2:end-1),q_a(2:end-1),v_p(2:end-1)); % (2:end-1) because q has been extended by boundary condition
%             this.debug = [this.debug,Fi];
            %Fa/x(idx.q_a(1));
%             p.A_a/p.rho*Fa/x(idx.v_p(25));
            gi = p.rho*g;
            ga = p.rho*g;
            gp = p.rho_p*g;
            

            %% right-hand side          
            % rhs
            dx = zeros(size(x));
            
            dx(idx.q_i) = - p.A_i/p.rho*(dpi+Fi+gi);
            if this.useHeaveLock == true
                dx(idx.q_i(1)-1+this.positionHeaveLock) = 0*dqi(this.positionHeaveLock);%10*(q_HeaveLock-q_i(this.positionHeaveLock));
            end
            dx(idx.v_p) = - 1/p.rho_p*(dpp + Fp + p.kp*v_p(2:end-1) + gp);
            dx(idx.q_a) = - p.A_a/p.rho*(dpa+Fa+ga) ;
            dx(idx.p_p) = p.E*( -dvp + p.fi*dx(idx.p_i) - p.fa*dx(idx.p_a));
            
            % solve matrix*[dp_a/dt; dp_i/dt] = [RHSp_a; RHSp_i] at every
            % point (i.e. component)
            RHSp_i = -p.beta_i./p.A_i.*dqi;   % ToDo: consider rho?
            RHSp_a = -p.beta_a./p.A_a.*dqa;

            if this.coupling_pi_pa
                %betaamean = mean(p.beta_a);
                %betaimean = mean(p.beta_i);
                %matrix = [1, betaamean*p.Aa_pi; betaimean*p.Ai_pa, 1];
                matrix = [1, p.beta_a*p.Aa_pi; p.beta_i*p.Ai_pa, 1]; % if betas scalar
            else
                matrix = eye(2);  % if p_i and p_a not coupled
            end
            hlp = matrix\[RHSp_a';RHSp_i'];
            dx(idx.p_a) = hlp(1,:)';
            dx(idx.p_i) = hlp(2,:)';
            
            % BHA acceleration:
            %dx(idx.v_BHA) = 1/p.m_BHA*(p.A_p*(p_bh-p_p(1))+p.A_i*(p_bh-p_i(1)));  % ToDo: minus frictional pressure drop as function of q
            if this.continuous_circulation
                % evaluate friction (just dirty)
                F_BHA = evaluateFrictionBHA(this,q_bit,q_bha,v_BHA);
                dx(idx.v_BHA) = 1/p.m_BHA*(p_bh*(p.A_bha-p.A_bit)-p.A_p*p_p(1)-(p.A_i-p.A_bit)*p_i(1)- (p.A_bha-p.A_p-p.A_i)*p_a(1))-g-0*F_BHA/p.rho_p;  % ToDo: minus frictional pressure drop as function of q
            else
                dx(idx.v_BHA) = 1/p.m_BHA*(p_bh*p.A_bha-p.A_p*p_p(1)-p.A_i*p_i(1)- (p.A_bha-p.A_p-p.A_i)*p_a(1))-g;  % ToDo: minus frictional pressure drop as function of q
            end
            if this.lump_pipe
            % ToDo: (too) simple, lumped model
            dx(idx.v_BHA) = -p.E*p.A_p/p.L/p.m_BHA*((x_bit-10)-(x_p(end)-0));
            end
            % Ulf's variant:
            %dx(idx.v_BHA) = 1/p.m_BHA*(-p.K*v_p(1)+p_a(1)*(p.A_p+p.A_i)-p_p(1)*p.A_p-p_i(1)*p.A_i) - g;    % -g ??
            %dx(idx.v_BHA) = 1/p.m_BHA*(p_bh*(p.A_bha)-p.A_p*p_p(1)-p.A_i*p_i(1)- (p.A_bha-p.A_p-p.A_i)*p_a(1))-g;
            
            % volume below bit:
            mdot = p.rho*(-q_bit-q_bha);
            Vdot = v_BHA*A_well;
            V_under_bit = x_bit*A_well;
            m_under_bit = p.rho*V_under_bit;
            % rho depends on p:
            %rho = p.rho*exp((p_bh-1e5)/p.beta);
            %dx(idx.p_bh) = (p.beta/rho)*((mdot*V_under_bit-Vdot*m_under_bit)/(V_under_bit^2));
            dx(idx.p_bh) = (p.beta_a(1)/p.rho)*((mdot*V_under_bit-Vdot*m_under_bit)/(V_under_bit^2));
            dx(idx.p_bh) = -p.beta_a(1)/V_under_bit*(q_bit+q_bha+v_BHA*(p.A_bha-p.A_bit));
            
            dx(idx.x_bit) = v_BHA;
            
            if ~this.continuous_circulation
            dx(idx.fluid_level_i) = (q_bit-q_pump)/p.A_i + (v_p(1)-v_p(end));
            end
            
            % tracking variables
            if this.track_pi_top
                this.pi_top_tracked = [this.pi_top_tracked, p_i(end)];
            end
            this.t_tracked = [this.t_tracked; t];
            this.q_bit_tracked = [this.q_bit_tracked; q_bit];
            this.q_bha_tracked = [this.q_bha_tracked; q_bha];
            this.q_HL_tracked = [this.q_HL_tracked; q_i(this.positionHeaveLock)];
            this.q_des_tracked = [this.q_des_tracked; q_des];
            this.u_tracked = [this.u_tracked, u_hl];
            
            
            if ~isempty(idx.x_p)
            dx(idx.x_p) = v_p(2:end-1);
            end
            
            if ~this.continuous_circulation 
            dx(idx.p_i(end)) = 0.5*p.rho*g*dx(idx.fluid_level_i);
            end
            
            if this.coupling_pi_pa
                dx(idx.p_a(end)) = (RHSp_a(end) - matrix(1,2)* dx(idx.p_i(end)))/matrix(1,1);
            end
            
            if this.useSimulink
                dx(idx.simulink) = dx_simulink;
            end
            
        end
        
        
        function [Fi, Fa, Fp] = evaluateFriction(this,q_i,q_a,v_p)
            % computes the frictional pressure drop (in the same units as dp/dx)
            p = this.param;
            % geometry constants
            apb = 1.4901;  % see tmp.m
            mu = p.rho*p.nu;
            K = mu*apb;
            n = 1;
            ri =sqrt(p.A_i/pi);
            A_well = p.A_i + p.A_p + p.A_a;
            rw = sqrt(A_well/pi);
            rh = (rw-ri)/2;   % /2????? factor 4 in Fanning vs. Dracy friction factors...
            
            v_i = q_i/p.A_i;
            v_a = q_a/p.A_a;
            
            v_i_eff = v_i - v_p;
            v_a_eff = v_a - .5*v_p;
            
            % ToDo: use effective velocities??
            v_i = v_i_eff;
            v_a = v_a_eff;
            
            
            
            Rei = 2*p.rho*v_i*ri/mu;
            %Rea = p.rho*v_a.^(1-n)*rh^n/(2^(n-3)*K);  % if n ~= 1
            Rea = p.rho*v_a*rh^n/(2^(n-3)*K);
            
            Fi = (16./Rei)*p.rho.*v_i.^2/ri;
            Fa = (16./Rea)*p.rho.*v_a.^2/rh;
            
            Fi = Fi*0.5;
            Fa = Fa*0.5;
            
            % sign "convention": positive flow <=> positive friction
            Fi = sign(v_i).*abs(Fi);
            Fa = sign(v_a).*abs(Fa);
            
            Fi(v_i==0) = 0;
            Fa(v_a==0) = 0;
            
            Fp = - (Fi + .5*Fa);  % friction on pipe in opposite direction of fluid friction
            
        end
        
        function F = evaluateFrictionBHA(this,q_i,q_a,v_p)
            % computes the frictional pressure drop (in the same units as dp/dx)
            p = this.param;
            % geometry constants
            apb = 1.4901;  % see tmp.m
            mu = p.rho*p.nu;
            K = mu*apb;
            n = 1;
            ri =sqrt(p.A_i/pi);
            A_well = p.A_i + p.A_p + p.A_a;
            rw = sqrt(A_well/pi);
            rBHA = sqrt(p.A_bha/pi);
            rh = (rw-rBHA)/2;   % /2????? factor 4 in Fanning vs. Dracy friction factors...
            
            v_i = q_i/p.A_i;
            v_a = q_a/p.A_p; % A gap
            
            v_i_eff = v_i - v_p;
            v_a_eff = v_a - .5*v_p;
            
            % ToDo: use effective velocities??
            v_i = v_i_eff;
            v_a = v_a_eff;
            
            
            
            Rei = 2*p.rho*v_i*ri/mu;
            %Rea = p.rho*v_a.^(1-n)*rh^n/(2^(n-3)*K);  % if n ~= 1
            Rea = p.rho*v_a*rh^n/(2^(n-3)*K);
            
            Fi = (16./Rei)*p.rho.*v_i.^2/ri;
            Fa = (16./Rea)*p.rho.*v_a.^2/rh;
            
            Fi = Fi*0.5;
            Fa = Fa*0.5;
            
            % sign "convention": positive flow <=> positive friction
            Fi = sign(v_i).*abs(Fi);
            Fa = sign(v_a).*abs(Fa);
            
            Fi(v_i==0) = 0;
            Fa(v_a==0) = 0;
            
            F = - (Fi + .5*Fa);  % friction on pipe in opposite direction of fluid friction
            
        end
        
        
        
        function x0 = getX0(this)
        % gets the initial condition such that the system is at
        % steady state for the given pump flow rate - not really at steady state in some cases...
            x0 = zeros(this.dim,1);
            idx = this.idx;
            p = this.param;
            deltaX = p.L/p.N;
            
            
            if this.continuous_circulation
                q0 = this.q_pump(0);
            else
                q0 = 0;
            end
            
            
%             dpchoke = 0;  % ToDo: pressure drop over choke
            dpchoke = (q0/(p.C_d*p.A_choke*this.choke_opening(0)))^2*p.rho/2 - 0.5*deltaX*p.g*p.rho;
            

            if ~this.pressuredropexceldata
                C_d = 0.95;
                A_well = p.A_i + p.A_p + p.A_a;
                A_gap = A_well - p.A_bha;
                dpbha = p.rho/(2*C_d^2*A_gap^2)*abs(q0)^2;
            else
                dpbha = 169*abs(q0)^1.25*1e5;
            end
            
            
            if ~this.pressuredropexceldata
                % Imsland's report:
                C_d = 0.95;
                A_bit_nozzles = p.A_p;  % ToDo
                dpbit = p.rho/(2*C_d^2*A_bit_nozzles^2)*abs(q0)^2;
            else
                dpbit = 90000*abs(q0)^2*1e5;
            end
             
            depth = ((p.N-0.5)*deltaX:(-deltaX):(0.5*deltaX))';
            from_bottom = (0:deltaX:((p.N-1)*deltaX))';
            g = 9.81;
            
            [Fi, Fa, Fp] = this.evaluateFriction(q0*ones(p.N-1,1),-q0*ones(p.N-1,1),zeros(p.N-1,1));
            
            
            pa = 1e5 + dpchoke + [deltaX*flipud(cumsum(flipud(Fa)));0] + depth*p.g*p.rho;
            pbh = pa(1)+dpbha;
            pi = pbh + dpbit - [0;deltaX*cumsum(Fi)] - from_bottom*p.g*p.rho;
            if this.useHeaveLock
                dphl = (q0/(p.k_hl*p.u_hl_init))^2 - 0*p.rho*p.g*deltaX;
                this.debug = dphl;
                pi(this.positionHeaveLock+1:end) = pi(this.positionHeaveLock+1:end) + dphl;
            end
            pp = (1/p.A_p*(pbh*p.A_bha-p.A_i*pi(1) - (p.A_bha-p.A_p-p.A_i)*pa(1))) ...
                - p.m_BHA/p.A_p*p.g - from_bottom*p.rho_p*p.g - [0;deltaX*cumsum(Fp)];
            
            
            x0(idx.p_i) = pi;
            x0(idx.q_i) = q0;
            x0(idx.p_p) = pp;
            x0(idx.v_p) = 0;
            x0(idx.p_a) = pa;
            x0(idx.q_a) = -q0;
            x0(idx.v_BHA) = 0;
            x0(idx.p_bh) = pbh;  
            x0(idx.x_bit) = 10; % bit 10m above ground
            x0(idx.q_bit) = q0;
            x0(idx.q_BHA) = -q0;
            x0(idx.fluid_level_i) = p.L;
            
            if this.useSimulink
                x0(idx.simulink) = zeros(length(idx.simulink),1);
            end
            
            x0(idx.x_p) = 0;   % ToDo
        end
          
        %% plotting
        function plot_pi(this,t,x,offset)
            if nargin < 4
                offset = true;
            end
            
            if offset
                xx = (x(this.idx.p_i,:) - repmat(x(this.idx.p_i,1),1,length(t)))';
            else
                xx = x(this.idx.p_i,:)';
            end
            figure;
            dx = this.param.L/this.param.N;
            X = 0.5*dx:dx:(this.param.L-0.5*dx);
            surf(X,t,xx/1e5,'LineStyle','none')
            %mesh(X,t,xx/1e5)
            xlabel('x [m]')
            ylabel('t [s]')
            zlabel('p_i [bar]')
        end
        
        
        function plot_qi(this,t,x,offset)
            if nargin < 4
                offset = false;
            end
            
            if offset
                xx = (x(this.idx.q_i,:) - repmat(x(this.idx.q_i,1),1,length(t)))';
            else
                xx = x(this.idx.q_i,:)';
            end
            figure;
            dx = this.param.L/this.param.N;
            X = dx:dx:(this.param.L-dx);
            surf(X,t,xx,'LineStyle','none')
            %mesh(X,t,xx)
            xlabel('x [m]')
            ylabel('t [s]')
            zlabel('q_i [m^3/s]')
        end
        
        function plot_pp(this,t,x,offset)
            if nargin < 4
                offset = true;
            end
            
            if offset
                xx = (x(this.idx.p_p,:) - repmat(x(this.idx.p_p,1),1,length(t)))';
            else
                xx = x(this.idx.p_p,:)';
            end
            figure;
            dx = this.param.L/this.param.N;
            X = 0.5*dx:dx:(this.param.L-0.5*dx);
            surf(X,t,xx/1e5,'LineStyle','none')
            %mesh(X,t,xx/1e5)
            xlabel('x [m]')
            ylabel('t [s]')
            zlabel('p_p [bar]')
        end
        
        function plot_vp(this,t,x,offset)
            if nargin < 4
                offset = false;
            end
            
            if offset
                xx = (x(this.idx.v_p,:) - repmat(x(this.idx.v_p,1),1,length(t)))';
            else
                xx = x(this.idx.v_p,:)';
            end
            figure;
            dx = this.param.L/this.param.N;
            X = dx:dx:(this.param.L-dx);
            surf(X,t,xx,'LineStyle','none')
            %mesh(X,t,xx)
%             colormap jet
%             colormap cool
            xlabel('x [m]')
            ylabel('t [s]')
            zlabel('v_p [m/s]')
        end
        
        
        function plot_xp(this,t,x,offset)
            if nargin < 4
                offset = false;
            end
            
            if offset
                xx = (x(this.idx.x_p,:) - repmat(x(this.idx.x_p,1),1,length(t)))';
            else
                xx = x(this.idx.x_p,:)';
            end
            figure;
            dx = this.param.L/this.param.N;
            X = dx:dx:(this.param.L-dx);
            surf(X,t,xx,'LineStyle','none')
            %mesh(X,t,xx)
            colormap parula  % the matlab default value
%             colormap jet
%             colormap cool
            xlabel('x [m]')
            ylabel('t [s]')
            zlabel('x_p [m/s]')
        end
        
        
        function plot_pa(this,t,x,offset)
            if nargin < 4
                offset = true;
            end
            
            if offset
                xx = (x(this.idx.p_a,:) - repmat(x(this.idx.p_a,1),1,length(t)))';
            else
                xx = x(this.idx.p_a,:)';
            end
            figure;
            dx = this.param.L/this.param.N;
            X = 0.5*dx:dx:(this.param.L-0.5*dx);
            surf(X,t,xx/1e5,'LineStyle','none')
            % alpha(1);
            %mesh(X,t,xx/1e5)
            xlabel('x [m]')
            ylabel('t [s]')
            zlabel('p_a [bar]')
        end
        
        
        function plot_qa(this,t,x,offset)
            if nargin < 4
                offset = false;
            end
            
            if offset
                xx = (x(this.idx.q_a,:) - repmat(x(this.idx.q_a,1),1,length(t)))';
            else
                xx = x(this.idx.q_a,:)';
            end
            figure;
            dx = this.param.L/this.param.N;
            X = dx:dx:(this.param.L-dx);
            surf(X,t,xx,'LineStyle','none')
            %mesh(X,t,xx)
            xlabel('x [m]')
            ylabel('t [s]')
            zlabel('q_a [m^3/s]')
        end
        
        
        function plot_vbha(this,t,x,offset)
            if nargin < 4
                offset = false;
            end
            
            if offset
                xx = (x(this.idx.v_BHA,:) - repmat(x(this.idx.v_BHA,1),1,length(t)));
            else
                xx = x(this.idx.v_BHA,:);
            end
            figure;
            plot(t,xx);
            xlabel('t [s]');
            ylabel('v_{BHA} [m/s]');
        end
        
        function plot_xbit(this,t,x,offset)
            if nargin < 4
                offset = false;
            end
            
            if offset
                xx = (x(this.idx.x_bit,:) - repmat(x(this.idx.x_bit,1),1,length(t)));
            else
                xx = x(this.idx.x_bit,:);
            end
            figure;
            plot(t,xx);
            xlabel('t [s]');
            ylabel('x_{bit} [m]');
        end
        
        function plot_pbh(this,t,x,offset)
            if nargin < 4
                offset = false;
            end
            
            if offset
                xx = (x(this.idx.p_bh,:) - repmat(x(this.idx.p_bh,1),1,length(t)));
            else
                xx = x(this.idx.p_bh,:);
            end
            figure;
            plot(t,xx/1e5);
            hold all
            pmin = min(xx(round(end/2):end)/1e5);
            pmax = max(xx(round(end/2):end)/1e5);
            plot(t,repmat(pmin,length(t)),'r--',t,repmat(pmax,length(t)),'r--')
            text(2,pmax,['dp = ' num2str(pmax-pmin)],'VerticalAlignment','top')
            xlabel('t [s]');
            ylabel('p_{bh} [bar]');
        end
        
        function plot_qbit(this,t)
            figure;
            plot(t,-this.q_bit_tracked.*6e4,t,-this.q_HL_tracked.*6e4,t,this.q_bha_tracked.*6e4);
            xlabel('t [s]');
            ylabel('q [l/min]');
            legend('-q_{bit}','-q_{HL}','q_{bha}')
        end
        
        function init_idx(this)
            N = this.param.N;
            
            last = 0;
            
            % distributed variables
            N_p = N;
            N_q = N-1;
            idx.p_i = (last+1):(last+N_p); last=last+N_p;
            idx.q_i = (last+1):(last+N_q); last=last+N_q;
            idx.p_p = (last+1):(last+N_p); last=last+N_p;
            idx.v_p = (last+1):(last+N_q); last=last+N_q;
            idx.p_a = (last+1):(last+N_p); last=last+N_p;
            idx.q_a = (last+1):(last+N_q); last=last+N_q;
            % idx.x_p = (last+1):(last+N_q); last=last+N_q;  % pipe
            % displacement, might not be needed
            idx.x_p = [];
            
            % scalar variables
            idx.v_BHA = last+1; last=last+1;
            idx.p_bh = last+1; last=last+1;
            idx.x_bit = last+1; last=last+1;
            idx.q_bit = last+1; last=last+1;
            idx.q_BHA = last+1; last=last+1;
            idx.fluid_level_i = last+1;last=last+1;
            if this.useSimulink
                idx.simulink = (last+1):(last+this.states_simulink);last=last+this.states_simulink;
            end
            
            this.idx = idx;
            this.dim = last;
                       
        end      
    end   
end